@@ -7,11 +7,10 @@ from python_on_whales import docker
 import os
 import platform
 import datetime
+import pprint
 import time
-import secrets
-import socket
+import traceback
 import typing
-import re
 from jinja2 import Environment, FileSystemLoader
 from jinja2.exceptions import TemplateNotFound
 import mysql.connector
@@ -19,9 +18,9 @@ from mysql.connector import Error
 from pathlib import Path
 from wikibot3rd.wikiuser import WikiUser
 from mwdocker.logger import Logger
+from mwdocker.config import MwClusterConfig
 from mwdocker.html_table import HtmlTables
 from mwdocker.mariadb import MariaDB
-import pprint
 from lodstorage.lod import LOD
 from dataclasses import dataclass
 
@@ -80,93 +79,63 @@ class DBStatus():
     ok: bool
     msg:str
     ex:typing.Optional[Exception]=None
-    
+        
 class DockerApplication(object):
     '''
     MediaWiki Docker image
     '''
 
-    def __init__(self,user:str,password:str,name="mediawiki",version="1.35.7",container_name:str=None,extensionMap:dict={},wikiId:str=None,mariaDBVersion="10.9",smwVersion=None,logo=None,port=9080,sqlPort=9306,mySQLRootPassword=None,debug=False,verbose=True):
+    def __init__(self,
+                 config:MwClusterConfig,
+                 home:str=None):
         '''
         Constructor
         
         Args:
-            user(str): the initial sysop user to create
-            password(str): the initial sysop password to user
-            version(str): the  MediaWiki version to create docker applications for
-            container_name(str): the name of the container to be used
-            extensionMap(dict): a map of extensions to be installed
-            wikiId(str): the wikiId to create a py-3rdparty-mediawiki user for (if any)
-            sqlPort(int): the base port to be used for  publishing the SQL port (3306) for the docker applications
-            port(int): the port to be used for publishing the HTML port (80) for the docker applications
-            networkName(str): the name to use for the docker network to be shared by the cluster participants
-            mariaDBVersion(str): the Maria DB version to install as the SQL database provider for the docker applications
-            smwVersion(str): Semantic MediaWiki Version to be used (if any)
-            mySQLRootPassword(str): the mySQL root password to use for the database containers - if None a random password is generated
-            logo(str): URL of the logo to be used
-            debug(bool): if True debugging is enabled
-            verbose(bool): if True output is verbose
+            config: MwClusterConfig,
+            home: the home directory to use
         '''
-        # identifications
-        self.name=name
-        self.user=user
-        self.password=password
-        self.wikiId=wikiId
-        # extensions
-        self.extensionMap=extensionMap
-        # Versions
-        self.version=version
-        self.fullVersion=f"MediaWiki {self.version}"
-        self.smwVersion=smwVersion
-        self.underscoreVersion=version.replace(".","_")
-        self.shortVersion=self.getShortVersion()
-        # container names
-        if container_name is None:
-            self.container_name=f"mw{self.underscoreVersion}"
-        else:
-            self.container_name=container_name
+        self.config=config
         # branch as need for git clone e.g. https://gerrit.wikimedia.org/g/mediawiki/extensions/MagicNoCache
-        self.branch=f"REL{self.getShortVersion('_')}"
-        self.mariaDBVersion=mariaDBVersion
+        self.branch=f"REL{self.config.getShortVersion('_')}"
         self.composerVersion=1
-        if self.shortVersion>="139":
+        if self.config.shortVersion>="139":
             self.composerVersion=2
-        # hostname and ports
-        self.hostname=socket.getfqdn()
-        if self.hostname=="1.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.ip6.arpa":
-            self.hostname="localhost"
-        self.port=port
-        self.url=f"http://{self.hostname}:{self.port}"
-        self.sqlPort=sqlPort
-        self.logo=logo
-        # debug and verbosity
-        self.debug=debug
-        self.verbose=verbose
-        # passwords
-        password_length = 13
-        if mySQLRootPassword is None:
-            self.mySQLRootPassword=secrets.token_urlsafe(password_length)
-        else:
-            self.mySQLRootPassword=mySQLRootPassword
-        self.mySQLPassword=secrets.token_urlsafe(password_length)
         # jinja and docker prerequisites
         self.env=self.getJinjaEnv()
+        if home is None:
+            home = str(Path.home())
+        self.dockerPath=f'{home}/.pymediawikidocker/{self.config.container_base_name}' 
+        os.makedirs(self.dockerPath,exist_ok=True)
+        
         self.getContainers()
         self.dbConn=None
         self.database="wiki"
-        self.host="localhost"
         self.dbUser="wikiuser"
         self.wikiUser=None
        
     @staticmethod 
-    def check()->str:
+    def check(debug:bool=False)->str:
+        """
+        check the docker environment
+        
+        Args:
+            debug(bool): if True show debug information
+            
+        Returns:
+            str: an error message or None
+        """
         errMsg=None
         if not docker.compose.is_installed():
-            errMsg="""docker composer up needs to be working
-            you might want to install https://github.com/docker/compose-cli
-            Compose v2 can be installed manually as a CLI plugin, 
-            by downloading latest v2.x release from https://github.com/docker/compose-cli/releases for your architecture and move into ~/.docker/cli-plugins/docker-compose
-"""
+            errMsg="""docker composer up needs to be working"""
+        os_path=os.environ["PATH"]
+        paths=["/usr/local/bin"]
+        for path in paths:
+            if os.path.islink(f"{path}/docker"):
+                if not path in os_path:
+                    os.environ["PATH"]=f"{os_path}{os.pathsep}{path}"
+                    if debug:
+                        print(f"""modified PATH from {os_path} to \n{os.environ["PATH"]}""")
         return errMsg
     
     def checkWiki(self,version_url:str)->bool:
@@ -178,7 +147,7 @@ class DockerApplication(object):
         try:
             html_tables=HtmlTables(version_url)
             tables=html_tables.get_tables("h2")
-            if self.debug:
+            if self.config.debug:
                 p = pprint.PrettyPrinter(indent=2)
                 p.pprint(tables)
             ok=ok and Logger.check_and_log("Special Version accessible ...", "Installed software" in tables)
@@ -186,10 +155,10 @@ class DockerApplication(object):
                 software=tables["Installed software"]
                 software_map,_dup=LOD.getLookup(software, "Product", withDuplicates=False)
                 mw_version=software_map["MediaWiki"]["Version"]
-                ok=ok and Logger.check_and_log_equal("Mediawiki Version",mw_version,"expected ",self.version)
+                ok=ok and Logger.check_and_log_equal("Mediawiki Version",mw_version,"expected ",self.config.version)
                 db_version_str=software_map["MariaDB"]["Version"]
                 db_version=MariaDB.getVersion(db_version_str)
-                ok=ok and Logger.check_and_log(f"Maria DB Version {db_version} fitting expected {self.mariaDBVersion}?",self.mariaDBVersion.startswith(db_version))
+                ok=ok and Logger.check_and_log(f"Maria DB Version {db_version} fitting expected {self.config.mariaDBVersion}?",self.config.mariaDBVersion.startswith(db_version))
                 pass
         except Exception as ex:
             ok=Logger.check_and_log(str(ex), False)
@@ -199,10 +168,7 @@ class DockerApplication(object):
         """
         get my container Name
         """
-        if self.container_name is None:
-            containerName=f"mw{self.underscoreVersion}{separator}{kind}{separator}1"
-        else:
-            containerName=f"{self.container_name}{separator}{kind}"
+        containerName=f"{self.config.container_base_name}{separator}{kind}"
         return containerName
     
     def getContainers(self):
@@ -232,9 +198,6 @@ class DockerApplication(object):
         resourcePath=os.path.realpath(f"{scriptdir}/resources")
         template_dir = os.path.realpath(f'{resourcePath}/templates')
         #print(f"jinja template directory is {template_dir}")
-        home = str(Path.home())
-        self.dockerPath=f'{home}/.pymediawikidocker/{self.container_name}' 
-        os.makedirs(self.dockerPath,exist_ok=True)
         env = Environment(loader=FileSystemLoader(template_dir))
         return env
     
@@ -248,9 +211,6 @@ class DockerApplication(object):
         self.execute("/tmp/update.sh")
         # add an initial sysop user as specified
         self.execute("/tmp/addSysopUser.sh")
-        # if wikiId is specified create a wikiuser
-        if self.wikiId is not None:
-            self.wikiUser=self.createWikiUser(store=True)
             
     def installExtensions(self):
         '''
@@ -264,31 +224,30 @@ class DockerApplication(object):
         '''
         self.execute("/root/addCronTabEntry.sh")
             
-            
-    def createWikiUser(self,store:bool=False):
+    def createWikiUser(self,wikiId:str=None,store:bool=False):
         '''
         create my wikiUser and optionally save it
         
         Args:
            store(bool): if True save my user data to the relevant ini File
         '''
-        if self.wikiId is None:
-            raise Exception("createWikiUser needs wikiId to be set but it is None")
+        if not wikiId:
+            wikiId=f"{self.config.container_base_name}"
         userDict={
-            "wikiId":f"{self.wikiId}",
-            "url": f"{self.url}",
-            "scriptPath": "",
-            "user": f"{self.user}",
+            "wikiId":f"{wikiId}",
+            "url": f"{self.config.base_url}",
+            "scriptPath": f"{self.config.script_path}",
+            "user": f"{self.config.user}",
             "email":"noreply@nouser.com",
-            "version": f"{self.fullVersion}",
-            "password": f"{self.password}"
+            "version": f"{self.config.fullVersion}",
+            "password": f"{self.config.password}"
         }
         wikiUser=WikiUser.ofDict(userDict,encrypted=False)
         if store:
             wikiUser.save()
         return wikiUser
     
-    def execute(self,command_str):
+    def execute(self,command_str:str):
         '''
         execute the given command string
         
@@ -296,22 +255,25 @@ class DockerApplication(object):
             command_str: str - a command string to be executed ...
         '''
         if self.mwContainer:
-            if self.verbose:
+            if self.config.verbose:
                 print(f"Executing docker command {command_str}")
             docker.execute(container=self.mwContainer.container,command=[command_str])
         else:
             mwContainerNameDash=self.getContainerName("mw", "-")
             mwContainerNameUnderscore=self.getContainerName("mw", "_")
-            errMsg=f"no mediawiki Container {mwContainerNameDash} or {mwContainerNameUnderscore} for {self.name} activated by docker compose\n- you might want to check the separator {self.mw_container_name_separator} for your platform {platform.system()}"
+            errMsg=f"no mediawiki Container {mwContainerNameDash} or {mwContainerNameUnderscore} for {self.name} activated by docker compose\n- you might want to check the separator character used for container names for your platform {platform.system()}"
             raise Exception(f"{errMsg}")
     
     def close(self):
+        """
+        close the database
+        """
         self.dbClose()
     
     def sqlQuery(self,query):
-        '''
+        """
         run the given SQL query
-        '''
+        """
         if self.dbConn and self.dbConn.is_connected():
             cursor = self.dbConn.cursor()
             cursor.execute(query)
@@ -319,19 +281,19 @@ class DockerApplication(object):
             cursor.close()
             return rows
         else:
-            if self.verbose:
-                print (f"Connection to {self.database} on {self.host} with user {self.dbUser} not established" )
+            if self.config.verbose:
+                print (f"Connection to {self.database} on {self.config.host} with user {self.dbUser} not established" )
             return None
         
     def dbClose(self):
-        '''
+        """
         close the database connection
-        '''
+        """
         if self.dbConn and self.dbConn.is_connected():
             self.dbConn.close()
         
     def dbConnect(self,timeout:int=10):
-        '''
+        """
         connect to the database and return the connection
         
         Args:
@@ -339,25 +301,25 @@ class DockerApplication(object):
             
         Returns:
             the connection
-        '''
+        """
         if self.dbConn is None:
             try:
-                self.dbConn = mysql.connector.connect(host=self.host,
+                self.dbConn = mysql.connector.connect(host=self.config.host,
                                  database=self.database,
                                  user=self.dbUser,
-                                 port=self.sqlPort,
-                                 password=self.mySQLPassword,
+                                 port=self.config.sqlPort,
+                                 password=self.config.mySQLPassword,
                                  connection_timeout=timeout)
         
             except Error as e :
                 errMsg=str(e)
-                print (f"Connection to {self.database} on {self.host} with user {self.dbUser} failed error: {errMsg}" )
+                print (f"Connection to {self.database} on {self.config.host} with user {self.dbUser} failed error: {errMsg}" )
                 if "Access denied" in errMsg:
                     raise e
         return self.dbConn
     
     def doCheckDBConnection(self,dbStatus:DBStatus,timeout:int=10):
-        '''
+        """
         check the database connection of this application
         
         Args:
@@ -365,17 +327,17 @@ class DockerApplication(object):
             
         Returns:
             DBStatus
-        '''          
+        """          
         dbStatus.attempts+=1
         self.dbConnect(timeout=timeout)
         if self.dbConn and self.dbConn.is_connected():
             rows=self.sqlQuery("select database();")
             dbStatus.ok=True
-            if self.verbose:
+            if self.config.verbose:
                 print (f"{dbStatus.msg} established database returns: {rows[0]}")     
     
     def checkDBConnection(self,timeout:float=10,initialSleep:float=2.5,maxTries:int=7)->DBStatus:
-        '''
+        """
         check database connection with retries
         
         Args:
@@ -386,10 +348,10 @@ class DockerApplication(object):
             
         Returns:
             dbStatus: the status
-        ''' 
-        conn_msg=f"SQL-Connection to {self.database} on {self.host} port {self.sqlPort} with user {self.dbUser}"
+        """ 
+        conn_msg=f"SQL-Connection to {self.database} on {self.config.host} port {self.config.sqlPort} with user {self.dbUser}"
         dbStatus=DBStatus(attempts=0,ok=False,msg=conn_msg,max_tries=maxTries)
-        if self.debug:
+        if self.config.debug:
             print (f"Trying {dbStatus.msg} with max {maxTries} tries and {timeout}s timeout per try - initial sleep {initialSleep}s")
         time.sleep(initialSleep)
         sleep=0.5
@@ -397,63 +359,70 @@ class DockerApplication(object):
             try:
                 self.doCheckDBConnection(dbStatus,timeout=timeout)
                 if not dbStatus.ok:
-                    if self.verbose:
+                    if self.config.verbose:
                         print(f"Connection attempt #{dbStatus.attempts}/{dbStatus.max_tries} failed will retry in {sleep} secs" )   
                     # wait before trying
                     time.sleep(sleep)
                     sleep=sleep*2
             except Exception as ex:
                 dbStatus.ex=ex
-                if self.verbose:
-                    print(f"Connection attempt #{dbStatus.attempts} failed with exception - will not retry ...")
+                if self.config.verbose:
+                    print(f"Connection attempt #{dbStatus.attempts} failed with exception {str(ex)} - will not retry ...")
+                if self.config.debug:
+                    print(traceback.format_exc())
                 break
         return dbStatus
     
-    def generate(self,templateName:str,targetPath:str,**kwArgs):
+    def optionalWrite(self,targetPath:str,content:str,overwrite:bool=False):
+        """
+        optionally Write the modified content to the given targetPath
+        
+        Args:
+            targetPath(str): the path to write the content to
+            content(str): the content to write
+            overwrite(bool): if True overwrite the existing content
+        """
+        if not overwrite and os.path.isfile(targetPath):
+            if self.config.verbose:
+                print(f"{targetPath} already exists!")
+            return
+        with open(targetPath, "w") as targetFile:
+                targetFile.write(content)
+    
+    def generate(self,templateName:str,targetPath:str,overwrite:bool=False,**kwArgs):
         '''
         generate file at targetPath using the given templateName
         
         Args:
             templateName(str): the Jinja2 template to use
             targetPath(str): the path to the target file
+            overwrite(bool): if True overwrite existing files
             kwArgs(): generic keyword arguments to pass on to template rendering
-        
         '''
         try:
             template = self.env.get_template(templateName)
             timestamp=datetime.datetime.now().isoformat()
-            content=template.render(mwVersion=self.version,mariaDBVersion=self.mariaDBVersion,port=self.port,sqlPort=self.sqlPort,smwVersion=self.smwVersion,timestamp=timestamp,**kwArgs)
-            with open(targetPath, "w") as targetFile:
-                targetFile.write(content)
+            content=template.render(mwVersion=self.config.version,mariaDBVersion=self.config.mariaDBVersion,port=self.config.port,sqlPort=self.config.sqlPort,smwVersion=self.config.smwVersion,timestamp=timestamp,**kwArgs)
+            self.optionalWrite(targetPath, content, overwrite)
+
         except TemplateNotFound:
-            print(f"no template {templateName} for {self.name} {self.version}")     
-        
-    def getShortVersion(self,separator=""):
-        '''
-        get my short version e.g. convert 1.27.7 to 127
-        
-        Returns:
-            str: the short version string
-        '''
-        versionMatch=re.match("(?P<major>[0-9]+)\.(?P<minor>[0-9]+)",self.version)
-        shortVersion=f"{versionMatch.group('major')}{separator}{versionMatch.group('minor')}"
-        return shortVersion
+            print(f"no template {templateName} for {self.config.name} {self.config.version}")     
     
     def getComposerRequire(self):
         '''
         get the json string for the composer require e.g. composer.local.json
         '''
         requires=[]
-        for ext in self.extensionMap.values():
+        for ext in self.config.extensionMap.values():
             if hasattr(ext,"composer"):
                 # get the composer statement
                 composer=ext.composer
-                requires.append(ext.composer)
+                requires.append(composer)
         indent="     "
         delim="" if len(requires)==0 else ",\n"
         requireList=""
-        if self.smwVersion:
-            requireList+=f'{indent}"mediawiki/semantic-media-wiki": "~{self.smwVersion}"{delim}'
+        if self.config.smwVersion:
+            requireList+=f'{indent}"mediawiki/semantic-media-wiki": "~{self.config.smwVersion}"{delim}'
         for i,require in enumerate(requires):
             delim="" if i>=len(requires)-1 else ",\n"
             requireList+=f"{indent}{require}{delim}"
@@ -464,7 +433,7 @@ class DockerApplication(object):
 }}"""
         return requireJson
         
-    def genComposerRequire(self,composerFilePath):
+    def genComposerRequire(self,composerFilePath,overwrite:bool=False):
         '''
         gen the composer.local.json require file
         
@@ -472,39 +441,53 @@ class DockerApplication(object):
             composerFilePath(str): the name of the file to generate
         '''
         requireJson=self.getComposerRequire()
-        with open(composerFilePath, "w") as composerFile:
-                composerFile.write(requireJson)
-                
+        self.optionalWrite(composerFilePath, requireJson, overwrite)
              
-    def generateAll(self):
+    def generateAll(self,overwrite:bool=False):
         '''
         generate all files needed for the docker handling
-        '''
-        self.generate("mwDockerfile",f"{self.dockerPath}/Dockerfile",composerVersion=self.composerVersion)
-        self.generate("mwCompose.yml",f"{self.dockerPath}/docker-compose.yml",mySQLRootPassword=self.mySQLRootPassword,mySQLPassword=self.mySQLPassword,container_name=self.container_name)
-        self.generate(f"mwLocalSettings{self.shortVersion}.php",f"{self.dockerPath}/LocalSettings.php",mySQLPassword=self.mySQLPassword,hostname=self.hostname,extensions=self.extensionMap.values(),mwShortVersion=self.shortVersion,logo=self.logo)
-        self.generate(f"mwWiki{self.shortVersion}.sql",f"{self.dockerPath}/wiki.sql")
-        self.generate(f"addSysopUser.sh",f"{self.dockerPath}/addSysopUser.sh",user=self.user,password=self.password)
-        self.generate(f"installExtensions.sh",f"{self.dockerPath}/installExtensions.sh",extensions=self.extensionMap.values(),branch=self.branch)
-        self.genComposerRequire(f"{self.dockerPath}/composer.local.json")
-        for fileName in ["addCronTabEntry.sh","startRunJobs.sh","initdb.sh","update.sh","phpinfo.php","upload.ini","plantuml.sh"]:
-            self.generate(f"{fileName}",f"{self.dockerPath}/{fileName}")
         
+        Args:
+            overwrite(bool): if True overwrite the existing files
+        '''
+        self.generate("mwDockerfile",f"{self.dockerPath}/Dockerfile",composerVersion=self.composerVersion,overwrite=overwrite)
+        self.generate("mwCompose.yml",f"{self.dockerPath}/docker-compose.yml",mySQLRootPassword=self.config.mySQLRootPassword,mySQLPassword=self.config.mySQLPassword,container_base_name=self.config.container_base_name,overwrite=overwrite)
+        self.generate(f"mwLocalSettings{self.config.shortVersion}.php",f"{self.dockerPath}/LocalSettings.php",mySQLPassword=self.config.mySQLPassword,hostname=self.config.host,extensions=self.config.extensionMap.values(),mwShortVersion=self.config.shortVersion,logo=self.config.logo,overwrite=overwrite)
+        self.generate(f"mwWiki{self.config.shortVersion}.sql",f"{self.dockerPath}/wiki.sql",overwrite=overwrite)
+        self.generate(f"addSysopUser.sh",f"{self.dockerPath}/addSysopUser.sh",user=self.config.user,password=self.config.password,overwrite=overwrite)
+        self.generate(f"installExtensions.sh",f"{self.dockerPath}/installExtensions.sh",extensions=self.config.extensionMap.values(),branch=self.branch,overwrite=overwrite)
+        self.genComposerRequire(f"{self.dockerPath}/composer.local.json",overwrite=overwrite)
+        for fileName in ["addCronTabEntry.sh","startRunJobs.sh","initdb.sh","update.sh","phpinfo.php","upload.ini","plantuml.sh"]:
+            self.generate(f"{fileName}",f"{self.dockerPath}/{fileName}",overwrite=overwrite)
+        
+    def down(self,forceRebuild:bool=False):
+        """
+        run docker compose down
+        
+        see https://docs.docker.com/engine/reference/commandline/compose_down/
+        and https://gabrieldemarmiesse.github.io/python-on-whales/sub-commands/compose/#down
+        
+        """
+        # change directory so that docker CLI will find the relevant dockerfile and docker-compose.yml
+        if self.config.verbose:
+            print(f"running docker compose down for {self.config.container_base_name} {self.config.version} docker application ...")
+        os.chdir(self.dockerPath)
+        docker.compose.down(volumes=forceRebuild) 
         
     def up(self,forceRebuild:bool=False):
-        '''
+        """
         start this docker application
         
         Args: 
             forceRebuild(bool): if true stop and remove the existing containers
-        '''            
-        if self.verbose:
-            print(f"starting {self.name} {self.version} docker application ...")
+        """            
+        if self.config.verbose:
+            print(f"starting {self.config.container_base_name} {self.config.version} docker application ...")
         if forceRebuild:
             for docker_container in [self.dbContainer,self.mwContainer]:
                 if docker_container is not None:
                     container=docker_container.container
-                    if self.verbose:
+                    if self.config.verbose:
                         print(f"stopping and removing container {container.name}")
                     container.stop()
                     container.remove()
@@ -516,6 +499,27 @@ class DockerApplication(object):
             docker.compose.build()
         # run docker compose up
         # this might take a while e.g. downloading
-        docker.compose.up(detach=True)    
-        self.getContainers()
-            
+        # run docker compose up
+        docker.compose.up(detach=True,force_recreate=forceRebuild)      
+        return self.getContainers()
+        
+    def start(self,forceRebuild:bool=False,withInitDB=True):
+        """
+        start my containers
+        
+        Args:
+            forceRebuild(bool): if True force rebuilding
+            withInitDB(bool): if True intialize my database
+        """         
+        self.up(forceRebuild=forceRebuild) 
+        if withInitDB:
+            if self.config.verbose:
+                print("Initializing MediaWiki SQL tables")
+            dbStatus=self.checkDBConnection()
+            if dbStatus.ok:
+                # first install extensions
+                self.installExtensions()
+                # then create and fill database and update it
+                self.initDB()
+                # then run startUp scripts
+                self.startUp()
\ No newline at end of file
